// Copyright (c) 2003-2010 Christopher M. Kohlhoff
// Modifications (c) 2011 by Guilherme Maciel Ferreira / Intra2net AG
//           and (c) 2014 by Christian Herdtweck / Intra2net AG
//
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)
#include "host/messagepayload.h"

#include <algorithm>
#include <limits>

#include <logfunc.hpp>

#include "boost_assert_handler.h"

using namespace std;
using I2n::Logger::GlobalLogger;

//-----------------------------------------------------------------------------
// MessagePayload
//-----------------------------------------------------------------------------

/**
 * @brief Parameterized constructor.
 *
 * @param payload_size_in_bytes the size of the payload (i.e. the internal buffer)
 */
MessagePayload::MessagePayload(
        const std::size_t &payload_size_in_bytes
) :
    PayloadSizeInBytes( payload_size_in_bytes ),
    Payload( payload_size_in_bytes )
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );

    fill( Payload.begin(), Payload.end(), 0 );
}

/**
 * @brief Copy constructor.
 *
 * @param other The object from where to get the message data.
 */
MessagePayload::MessagePayload(
        const MessagePayload &other
) :
    PayloadSizeInBytes( other.PayloadSizeInBytes ),
    Payload( other.PayloadSizeInBytes )
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );
    BOOST_ASSERT( other.Payload.capacity() == other.PayloadSizeInBytes );

    copy( other.Payload.begin(), other.Payload.end(), Payload.begin() );

    BOOST_ASSERT( PayloadSizeInBytes == other.PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() >= other.Payload.size() );
    BOOST_ASSERT( equal( Payload.begin(), Payload.end(), other.Payload.begin() ) );
    BOOST_ASSERT( equal( other.Payload.begin(), other.Payload.end(), Payload.begin() ) );
}

/**
 * @brief Destructor.
 */
MessagePayload::~MessagePayload()
{
    // The Payload is automatically deleted when the smart pointer's scope ends
}

/**
 * @brief Assignment operator.
 *
 * @param other The object from where to get the message data.
 *
 * @return This object.
 */
MessagePayload& MessagePayload::operator=( const MessagePayload &other )
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );
    BOOST_ASSERT( other.Payload.capacity() == other.PayloadSizeInBytes );
    BOOST_ASSERT( !other.Payload.empty() );

    // checking for self-assignment (i.e. object1 = object1)
    if ( &other == this )
    {
        return *this;
    }

    size_t new_payload_size_in_bytes = other.PayloadSizeInBytes;
    size_t old_payload_size_in_bytes = PayloadSizeInBytes;
    if ( new_payload_size_in_bytes > old_payload_size_in_bytes )
    {
        Payload.reserve( new_payload_size_in_bytes );
        PayloadSizeInBytes = new_payload_size_in_bytes;
    }

    copy( other.Payload.begin(), other.Payload.end(), Payload.begin() );

    BOOST_ASSERT( PayloadSizeInBytes == other.PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() >= other.Payload.size() );
    BOOST_ASSERT( equal( Payload.begin(), Payload.end(), other.Payload.begin() ) );
    BOOST_ASSERT( equal( other.Payload.begin(), other.Payload.end(), Payload.begin() ) );

    return *this;               //lint !e429
}

/**
 * @brief Array Subscript Operator.
 *
 * The element access operator to provide access syntax like regular
 * arrays. This version is used when this object is constant.
 *
 * @return A non-modifiable reference to the object indexed by the offset.
 */
const uint8_t& MessagePayload::operator[]( size_t offset ) const
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( offset < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );

    ptrdiff_t index = static_cast<ptrdiff_t>( offset );
    return Payload[ index ];
}

/**
 * @brief Array Subscript Operator.
 *
 * The element access operator to provide access syntax like regular
 * arrays. This version is used by non-const objects.
 *
 * @return A modifiable reference to the object indexed by the offset.
 */
uint8_t& MessagePayload::operator[]( size_t offset )
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( offset < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );

    ptrdiff_t index = static_cast<ptrdiff_t>( offset );
    return Payload[ index ];
} //lint !e1762

/**
 * @brief Get a pointer to the underlying array.
 *
 * @return a pointer to the array encapsulated by this class.
 */
uint8_t* MessagePayload::get()
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );

    // Vector containers have their elements stored in contiguous storage
    // locations, which means that their elements can be accessed not only
    // using iterators but also using offsets on regular pointers to elements.
    return static_cast<uint8_t *>( &Payload[ 0 ] );
}

/**
 * @brief Adds extra bytes in the MessagePayload.
 *
 * @details If there is no more space available in the internal buffer, extra
 * space is allocated to hold the additional bytes.
 *
 * @param extra_payload_data The data bytes to append at the end of the current
 * payload.
 * @param extra_payload_size_in_bytes The amount of bytes in @a extra_payload_data
 * to store at the @c MessagePayload.
 *
 * @return void.
 */
void MessagePayload::append(
        const uint8_t *extra_payload_data,
        const size_t extra_payload_size_in_bytes
)
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );
    BOOST_ASSERT( extra_payload_data != NULL );

    if ( 0 < extra_payload_size_in_bytes )
    {
        BOOST_ASSERT( 0 < extra_payload_size_in_bytes );

        size_t old_payload_size_in_bytes = PayloadSizeInBytes;
        size_t new_payload_size_in_bytes = old_payload_size_in_bytes + extra_payload_size_in_bytes;
        PayloadSizeInBytes = new_payload_size_in_bytes;

        Payload.insert( Payload.end(),
                        &extra_payload_data[ 0 ],
                        &extra_payload_data[ extra_payload_size_in_bytes ]
        );

#ifndef BOOST_DISABLE_ASSERTS
        BOOST_ASSERT( PayloadSizeInBytes == ( old_payload_size_in_bytes + extra_payload_size_in_bytes ) );

        uint8_t *extra_data_begin = &Payload[ old_payload_size_in_bytes ];
        uint8_t *extra_data_end = &Payload[ old_payload_size_in_bytes + extra_payload_size_in_bytes ];
        BOOST_ASSERT( equal( extra_data_begin, extra_data_end, extra_payload_data ) );
#endif
    }
}

/**
 * @brief Retrieve 1 bit from the payload buffer.
 *
 * @param byte_index The index of the byte.
 * @param bit_index The index of bit within the byte. The indexes are as follow:
 * [0|1|2|3|4|5|6|7], with the index zero being the left most bit within the byte.
 *
 * @return The bit value stored at @a bit_index.
 */
bool MessagePayload::decode1(
        const int byte_index,
        const int bit_index
) const
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );
    const int max_bit_index = 7;
    BOOST_ASSERT( ( 0 <= byte_index ) && ( byte_index < static_cast<int>(PayloadSizeInBytes) ) );
    BOOST_ASSERT( ( 0 <= bit_index ) && ( bit_index <= max_bit_index ) );

    int shift_amount = max_bit_index - bit_index;
    uint8_t bit_mask = static_cast<uint8_t>( 0x01U << shift_amount );
    bool value = static_cast<bool>( Payload[ byte_index ] & bit_mask );

    return value;
}

/**
 * @brief Store 1 bit in the payload buffer.
 *
 * @param byte_index The index of the byte.
 * @param bit_index The index of bit within the byte.
 * @param value The bit value to store at @a bit_index.
 *
 * @return void
 */
void MessagePayload::encode1(
        const int byte_index,
        const int bit_index,
        const bool value
)
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );
    const int max_bit_index = 7;
    BOOST_ASSERT( ( 0 <= byte_index ) && ( byte_index < static_cast<int>(PayloadSizeInBytes) ) );
    BOOST_ASSERT( ( 0 <= bit_index ) && ( bit_index <= max_bit_index ) );

    int shift_amount = max_bit_index - bit_index;
    uint8_t bit_mask = static_cast<uint8_t>( 0x01U << shift_amount );

    if ( value )
    {
        Payload[ byte_index ] = Payload[ byte_index ] | bit_mask;
    }
    else
    {
        bit_mask = static_cast<uint8_t>( ~bit_mask );
        Payload[ byte_index ] = Payload[ byte_index ] & bit_mask;
    }
}

/**
 * @brief Retrieve 16 bits from the payload buffer.
 *
 * @param left_byte_index The index of the left byte.
 * @param right_byte_index The index of the right byte.
 *
 * @return A concatenation of the byte indexed by @a left_byte_index with the byte
 * indexed by @a right_byte_index.
 */
uint16_t MessagePayload::decode16(
        const int left_byte_index,
        const int right_byte_index
) const
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );
    BOOST_ASSERT( ( 0 <= left_byte_index ) && ( left_byte_index < static_cast<int>(PayloadSizeInBytes) ) );
    BOOST_ASSERT( ( 0 <= right_byte_index ) && ( right_byte_index < static_cast<int>(PayloadSizeInBytes) ) );
    BOOST_ASSERT( left_byte_index < right_byte_index );
    BOOST_ASSERT( ( static_cast<size_t>( right_byte_index - left_byte_index ) + 1 ) == sizeof(uint16_t) );

    uint32_t value = static_cast<uint16_t>( Payload[ left_byte_index ] << 8 );
    value += static_cast<uint16_t>( Payload[ right_byte_index ] );

    BOOST_ASSERT( value <= numeric_limits<uint16_t>::max() );

    return static_cast<uint16_t>( value );
}

/**
 * @brief Store 16 bits in the payload buffer.
 *
 * @param left_byte_index The index of the left byte.
 * @param right_byte_index The index of the right byte.
 * @param value A 16-bit data to be saved in the bytes indexed by
 * @a left_byte_index and @a right_byte_index.
 *
 * @return void
 */
void MessagePayload::encode16(
        const int left_byte_index,
        const int right_byte_index,
        const uint16_t value
)
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );
    BOOST_ASSERT( ( 0 <= left_byte_index ) && ( left_byte_index < static_cast<int>(PayloadSizeInBytes) ) );
    BOOST_ASSERT( ( 0 <= right_byte_index ) && ( right_byte_index < static_cast<int>(PayloadSizeInBytes) ) );
    BOOST_ASSERT( left_byte_index < right_byte_index );
    BOOST_ASSERT( ( static_cast<size_t>( right_byte_index - left_byte_index ) + 1 ) == sizeof(uint16_t) );

    Payload[ left_byte_index ] = static_cast<uint8_t>( value >> 8 );
    Payload[ right_byte_index ] = static_cast<uint8_t>( value & 0xFF );
} //lint !e1762

/**
 * @brief Retrieve 32 bits from the payload buffer.
 *
 * @param first_byte_index The index of the first byte out of 4.
 * @param last_byte_index The index of the last byte out of 4.
 *
 * @return a concatenation of 4 bytes, from the byte indexed by
 * @a first_byte_index to the byte indexed by @a last_byte_index.
 */
uint32_t MessagePayload::decode32(
        const int first_byte_index,
        const int last_byte_index
) const
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );
    BOOST_ASSERT( ( 0 <= first_byte_index ) && ( first_byte_index < static_cast<int>(PayloadSizeInBytes) ) );
    BOOST_ASSERT( ( 0 <= last_byte_index ) && ( last_byte_index < static_cast<int>(PayloadSizeInBytes) ) );
    BOOST_ASSERT( first_byte_index < last_byte_index );
    BOOST_ASSERT( ( static_cast<size_t>( last_byte_index - first_byte_index ) + 1 ) == sizeof(uint32_t) );

    int current_byte_index = first_byte_index;

    uint64_t value = static_cast<uint32_t>( Payload[ current_byte_index ] ) << 24;
    value += static_cast<uint32_t>( Payload[ ++current_byte_index ] << 16 );
    value += static_cast<uint32_t>( Payload[ ++current_byte_index ] << 8 );
    value += static_cast<uint32_t>( Payload[ ++current_byte_index ] );

    BOOST_ASSERT( value <= numeric_limits<uint32_t>::max() );
    BOOST_ASSERT( current_byte_index == last_byte_index );

    return static_cast<uint32_t>( value );
}

/**
 * @brief Store 32 bits in the payload buffer.
 *
 * @param first_byte_index The index of the first byte out of 4.
 * @param last_byte_index The index of the last byte out of 4.
 * @param value A 32-bit data to be saved in the bytes indexed from
 * the @a first_byte_index to the @a last_byte_index.
 *
 * @return void
 */
void MessagePayload::encode32(
        const int first_byte_index,
        const int last_byte_index,
        const uint32_t value
)
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );
    BOOST_ASSERT( ( 0 <= first_byte_index ) && ( first_byte_index < static_cast<int>(PayloadSizeInBytes) ) );
    BOOST_ASSERT( ( 0 <= last_byte_index ) && ( last_byte_index < static_cast<int>(PayloadSizeInBytes) ) );
    BOOST_ASSERT( first_byte_index < last_byte_index );
    BOOST_ASSERT( ( static_cast<size_t>( last_byte_index - first_byte_index ) + 1 ) == sizeof(uint32_t) );

    int current_byte_index = first_byte_index;

    Payload[ current_byte_index ] = static_cast<uint8_t>( ( value >> 24 ) & 0xFF );
    Payload[ ++current_byte_index ] = static_cast<uint8_t>( ( value >> 16 ) & 0xFF );
    Payload[ ++current_byte_index ] = static_cast<uint8_t>( ( value >> 8 ) & 0xFF );
    Payload[ ++current_byte_index  ] = static_cast<uint8_t>( value & 0xFF );

    BOOST_ASSERT( current_byte_index == last_byte_index );
} //lint !e1762

void MessagePayload::encode_string( const int first_byte_index,
                                    const std::string value
)
{
    int length = static_cast<int>(value.length());

    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );
    BOOST_ASSERT( ( 0 <= first_byte_index ) && ( first_byte_index < static_cast<int>(PayloadSizeInBytes) ) );
    BOOST_ASSERT( first_byte_index +length <= static_cast<int>(PayloadSizeInBytes) );

    const uint8_t *data_converted = reinterpret_cast<const uint8_t *>(&value[0]);
    for (int index=0; index<length; ++index)
        Payload[first_byte_index + index] = data_converted[index];
}


/**
 * @brief Read/Extract all the data from the input stream @a is and stores it
 * in the payload buffer.
 *
 * @param is The stream object on which the action is performed.
 *
 * @return The stream object passed as parameter.
 */
istream& MessagePayload::read( istream &is )
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );

    char *payload_data_array = reinterpret_cast<char *>( &Payload[ 0 ] );

    (void) is.read( payload_data_array, static_cast<streamsize>( PayloadSizeInBytes ) );

    size_t data_received_in_bytes = static_cast<size_t>( is.gcount() );
    if ( data_received_in_bytes != PayloadSizeInBytes )
    {
        GlobalLogger.error() << "Expecting " << PayloadSizeInBytes
                << " bytes, but received " << is.gcount() << " bytes." << endl;
    }

    BOOST_ASSERT( Payload.size() == PayloadSizeInBytes );
    BOOST_ASSERT( !Payload.empty() );

    return is;
}

/**
 * @brief Write/Insert all the data to the output stream @a os that is stored in
 * the payload buffer.
 *
 * @param os The stream object on which the action is performed.
 *
 * @return The stream object passed as parameter.
 */
ostream& MessagePayload::write( ostream &os ) const
{
    BOOST_ASSERT( 0 < PayloadSizeInBytes );
    BOOST_ASSERT( Payload.capacity() == PayloadSizeInBytes );
    BOOST_ASSERT( Payload.size() == PayloadSizeInBytes );
    BOOST_ASSERT( !Payload.empty() );

    // Vector containers have their elements stored in contiguous storage
    // locations, which means that their elements can be accessed not only
    // using iterators but also using offsets on regular pointers to elements.
    const char *payload_data_array = reinterpret_cast<const char *>( &Payload[ 0 ] );

    (void) os.write( payload_data_array, static_cast<streamsize>( PayloadSizeInBytes ) );

    return os;
}

uint32_t MessagePayload::calc_checksum_part() const
{
    uint32_t sum = 0;

    std::vector<uint8_t>::const_iterator iter = Payload.begin();
    std::vector<uint8_t>::const_iterator end = Payload.end();
    while ( iter != end )
    {
        sum += (static_cast<uint8_t>( *iter++ ) << 8);
        if ( iter != end )
            sum += static_cast<uint8_t>( *iter++ );
    }

    return sum;
}
