/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.

 Christian Herdtweck, Intra2net AG 2015
 Based on an example in Boost Documentation (by Christopher M. Kohlhoff)
 and adaptation by Guilherme M. Ferreira
 */

#include "icmp/icmpdestinationunreachabledata.h"

#include <logfunc.hpp>
using I2n::Logger::GlobalLogger;


IcmpDestinationUnreachableData::IcmpDestinationUnreachableData(
                                                     const std::size_t size_arg)
    : IcmpData( size_arg )
{}

bool IcmpDestinationUnreachableData::match_echo_reply(const uint16_t identifier,
                      const uint16_t sequence_number) const
{ return false; }

bool IcmpDestinationUnreachableData::match_destination_unreachable(
                                           const uint16_t identifier,
                                           const uint16_t sequence_number) const
{
    return      identifier == get_icmp_identifier() &&
           sequence_number == get_icmp_sequence_number();
}


uint16_t IcmpDestinationUnreachableData::get_icmp_identifier() const
{ return get_icmp_request_data(5); }

uint16_t IcmpDestinationUnreachableData::get_icmp_sequence_number() const
{ return get_icmp_request_data(7); }

uint16_t IcmpDestinationUnreachableData::get_icmp_request_data(
                                                   const int data_offset) const
{
    // payload should be the original query, which is an IP packet.
    // first check wheter that IP packet contains an ICMP message at all
    bool is_icmp;
    int offset = 4;   // the 4 uninteresting bytes we need to skip
    if (IcmpData::raw_data[offset+0] == 4)   // IPv4
    {
        is_icmp = IcmpData::raw_data[offset+9] == 1;
        offset += 20;     // 20 byte for IPv4 header
    }
    else if (IcmpData::raw_data[offset+0] == 6)   // IPv6
    {
        is_icmp = IcmpData::raw_data[offset+6] == 58;
        offset += 40;     // 40 byte for IPv6 header
    }

    if ( !is_icmp )
        return static_cast<uint16_t>(-1);
    else
        // if it is an icmp message, then the icmp packet comes right after
        // the IP header. Inside the icmp packet we need bytes 5-8
        return IcmpData::raw_data.decode16(offset+data_offset,
                                           offset+data_offset+1);
}

/**
 * @brief Prints the destination unreachable messages.
 *
 * @return void
 */
void IcmpDestinationUnreachableData::print(
            const size_t &bytes_received,
            const boost::posix_time::ptime &time_packet_sent,
            const std::string &remote_address,
            const uint32_t ttl) const
{
    uint16_t sequence_number = get_icmp_sequence_number();

    GlobalLogger.info() << "Destination " << remote_address
         << " (icmp_seq=" << sequence_number << ") unreachable!"
         << std::endl;
}

// (created using vim -- the world's best text editor)

